#define DEBUG
// Globals:

GhostScrubber@ scrubber = null;
IGSlider@ slider = null;
string currentMap;
array<GhostScript@> temporaryGhosts;
bool haveCurrentRace = false;
// Auto hide slider:
int2 lastMousePos;
int lastMouseTime = 0;
// OSD button text state:
bool lastPaused = false;
// Precision scrubbing:
bool precisionMode = false;
vec2 precisionWindow;

// Records spectating
MwId forcedTarget;
bool targetVisible = false;
bool recordsFollow = false;
uint64 recordsStart = 0; // short bad delay, game script calls Spectator_SetForcedTarget_Ghost before setting UISequence
int lastStartTime = -1;

// bool shiftDown = false;

// Constants:

const int SCRUBBER_IDLE_TIME = 5000; // If mouse hasn't moved in this much time
const string SCRUBBER_PLAY_TEXT = Icons::Play; // Button text
const string SCRUBBER_PAUSE_TEXT = Icons::Pause; // Button text

const int RENDER_HIDE_TIME = -5000; // time to hide after start
const int RENDER_HIDE_DELAY = 200; // if speed < SPEED_THRESHOLD show if this much time has elapsed
const float SPEED_THRESHOLD = 1.0f; // ^

const uint64 RECORDS_STOP_DELAY = 100;

void AddCurrentRace() {
    if (Setting_Ghost_AddCurrentRace
        && !haveCurrentRace
        && ControlledPlayer !is null
        // important to add ghost as few times as possible because of 65 ghost bug:
        && (slider.Value * 1000.0) < cast<CSmScriptPlayer@>(ControlledPlayer.ScriptAPI).CurrentRaceTime) {
        auto ghost = PlaygroundScript.Ghost_RetrieveFromPlayer2(cast<CSmScriptPlayer@>(ControlledPlayer.ScriptAPI), Setting_Ghost_TruncateCPs);
        ghost.Nickname = Setting_Ghost_CurrentRaceName;
        AddTemporaryGhost(ghost);
        haveCurrentRace = true;
    }
}

void AddTemporaryGhost(CGameGhostScript@ ghost) {
    auto gs = GhostScript(ghost, true);
    gs.wantGhost = true;
    temporaryGhosts.InsertLast(gs);
}

void ClearTemporaryGhosts() {
    for (uint i = 0; i < temporaryGhosts.Length; i++) {
        temporaryGhosts[i].ReleaseGhost();
    }
    temporaryGhosts = {};
    haveCurrentRace = false;
}

void SetPrecisionMode(bool state) {
    precisionMode = state;
    slider.DisplayRange = precisionMode;

    const float WindowSize = Setting_Slider_PrecisionWindow;
    const float Value = slider.Value;
    auto min = Math::Max(0.f, Value - WindowSize / 2);
    auto max = Math::Min(scrubber.RaceLength / 1000.0, min + WindowSize);
    if (min > 0 && max - min < WindowSize) {
        min = Math::Max(0.f, max - WindowSize);
    }
    precisionWindow = vec2(min, max);

    // todo: pull out, settings, etc
    if (state) {
        slider.BorderColor = vec4(0.9, 1, 0.9, 0.75);
        slider.BackColor = vec4(0.8, 0.9, 0.8, 0.25);
        slider.ButtonColor = vec4(0.55, 0.66, 0.55, 0.9);
    } else {
        slider.BorderColor = vec4(1, 1, 1, 0.75);
        slider.BackColor = vec4(0.9, 0.9, 0.9, 0.25);
        slider.ButtonColor = vec4(0.66, 0.66, 0.66, 0.9);
    }
}

bool IsOSDVisible() {
    return (Time::Now - lastMouseTime) < SCRUBBER_IDLE_TIME;
}

bool IsFollowingRecord() {
    return Setting_Ghost_RecordsSpectate && recordsFollow;
}

// Update:

void Update(float dt) {
    // Current map detection
    if (currentMap.Length > 0 && (PlaygroundScript is null || PlaygroundScript.Map is null)) {
        currentMap = '';
        ClearTemporaryGhosts();
    } else if (PlaygroundScript !is null && PlaygroundScript.Map !is null && PlaygroundScript.Map.IdName != currentMap) {
        currentMap = PlaygroundScript.Map.IdName;
        ClearTemporaryGhosts();
    }

    auto player = ControlledPlayer;

    // Stopped following record / unpause at end of round
    if (IsFollowingRecord()
        && GameTerminal !is null
        && !(GameTerminal.UISequence_Current == CGamePlaygroundUIConfig::EUISequence::EndRound
            || GameTerminal.UISequence_Current == CGamePlaygroundUIConfig::EUISequence::UIInteraction)) {
        if (recordsStart == 0 || (Time::Now - recordsStart) > RECORDS_STOP_DELAY) {
            StopRecordsSpectate();
        }
        scrubber.Paused = false;
    }

    // Update scrubber
    if (Setting_Ghost_AddCurrentRace
        && !haveCurrentRace
        && player !is null
        && PlaygroundScript !is null
        && PlaygroundScript.DataFileMgr !is null // make sure ghosts can be added before updating time
        && !IsFollowingRecord()
        && !scrubber.Paused
        && !slider.IsMouseDown) {
        // Update race length for potential Current race
        scrubber.RaceLength = cast<CSmScriptPlayer@>(player.ScriptAPI).CurrentRaceTime;
    }
    scrubber.Update(dt);

    // attempt to remove "Current race" ghost before end of round replay
    if (scrubber.RaceTime >= scrubber.RaceLength) {
        ClearTemporaryGhosts();
        ResetSpectatorRace();
    }

    // Detect state of followed ghost from records (could use this to restart? not great either way)
    if (IsFollowingRecord() && PlaygroundScript !is null && PlaygroundScript.UIManager !is null) {
        bool visible = !PlaygroundScript.Ghost_IsReplayOver(forcedTarget);
        if (visible != targetVisible) {
            SetSpectatorTarget(visible);

            // Reset ghosts when target car is no longer visible (potentially a
            // bunch of wasted space at the end)
            if (!visible) {
                ResetSpectatorRace();
            }
            targetVisible = visible;
        }
    }

    // Hide slider when moving
    // todo: detect cam 7?
    if (Setting_OSD_Hide) {
        if (player !is null) {
            auto speed = cast<CSmScriptPlayer@>(player.ScriptAPI).Speed;
            auto time = cast<CSmScriptPlayer@>(player.ScriptAPI).CurrentRaceTime;

            if (UI::IsOverlayShown() || (time > RENDER_HIDE_DELAY && speed < SPEED_THRESHOLD)) {
                slider.Visible = true;
            } else {
                slider.Visible = time < RENDER_HIDE_TIME;
            }
        }
    } else {
        slider.Visible = true;
    }

    // Update slider button text
    if (lastPaused != scrubber.Paused) {
        slider.ButtonText = scrubber.Paused ? SCRUBBER_PAUSE_TEXT : SCRUBBER_PLAY_TEXT;
        lastPaused = scrubber.Paused;
    }


    // Display maximum seek value
    if (PlaygroundScript !is null) {
        slider.MarkerValue = PlaygroundScript.Now / 1000.0;
    }

    // Update slider
    slider.Value = scrubber.RaceTime / 1000.0;
    if (precisionMode) {
        slider.Min = precisionWindow.x;
        slider.Max = precisionWindow.y;
    } else {
        slider.Min = 0;
        slider.Max = scrubber.RaceLength / 1000.0;
    }
    slider.Update(dt);
}

// Input:

void Render() {
    if (PlaygroundScript !is null
        && PlaygroundScript.MapLoaded
        && GameTerminal !is null
        && (GameTerminal.UISequence_Current == CGamePlaygroundUIConfig::EUISequence::Playing
            || (IsFollowingRecord() && (GameTerminal.UISequence_Current == CGamePlaygroundUIConfig::EUISequence::EndRound
                || GameTerminal.UISequence_Current == CGamePlaygroundUIConfig::EUISequence::UIInteraction)))
        && IsOSDVisible()
        && IsInterfaceVisible()) {
        slider.Render();
    }
}

UI::InputBlocking OnMouseButton(bool down, int button, int x, int y) {
    lastMouseTime = Time::Now;
    lastMousePos = int2(x, y);

    // ignore mouse down events when interface hidden, but allow mouse up
    if (slider.Visible && (!down || IsInterfaceVisible())) {

        // probably better to just have drag to end as reset strategy
#if false
        // shift click resets (will allow CP-synced ghosts to work again)
        if (button == 0 && shiftDown && slider.PointInside(x, y)) {
            scrubber.Reset();
        } else
#endif
        // normal click
        if (button == 0 && slider.OnMouseButton(down, button, x, y)) {
            return UI::InputBlocking::Block;
        }
        // pause on right click inside slider
        else if (down && button == 1 && slider.PointInside(x, y)) {
            scrubber.Paused = !scrubber.Paused;
            return UI::InputBlocking::Block;
        }
    }

    return UI::InputBlocking::DoNothing;
}

UI::InputBlocking OnKeyPress(bool down, VirtualKey key) {
    if (key == Setting_Slider_PrecisionKey) {
        if (precisionMode != down) {
            SetPrecisionMode(down);
        }
    }

#if false
    if (key == VirtualKey::Shift) {
        shiftDown = down;
    }
#endif
    return UI::InputBlocking::DoNothing;
}

void OnMouseMove(int x, int y) {
    // occasional OnMouseMove events when mouse hasn't moved
    if (x != lastMousePos.x || y != lastMousePos.y) {
        lastMouseTime = Time::Now;
        lastMousePos = int2(x, y);
    }
    slider.OnMouseMove(x, y);
}

void OnSliderChange(float Value) {
    scrubber.SetRaceTime(int(Value * 1000));
    if (!IsFollowingRecord()) {
        AddCurrentRace();
    }
}

// Records Spectating
void StartRecordsSpectate() {
    log('Start records spectating');
    if (recordsFollow) {
        return;
    }
    recordsFollow = true;
    recordsStart = Time::Now;
    targetVisible = true; // assume
    ClearTemporaryGhosts();
    scrubber.RaceLength = 0;
    scrubber.lastStart = lastStartTime;
}

void StopRecordsSpectate() {
    log('Stop records spectating');
    if (!recordsFollow) { return; }
    recordsFollow = false;
    recordsStart = 0;
    Ghosts_SetStartTime(-1);
}


// From SpectatorBase_Client.Script.txt
	/* NB ClientUI.SpectatorForceCameraType:
	 * 1: Script has 2 states
	 *			Forced -> will act as ForcedTarget specifies
	 *			Default -> input can focus a player or allplayers
	 * 2: Free camera, player cannot use inputs
	 * 15: same as 1 but player can escape in free camera (script totaly loses control)

	 * After checking C++ with FlorentT it appears that
	 * 0 is Replay
	 * 1 is Follow
	 * 2 is Free
	 * 14 is FollowForced
	 * 15 is DontChange

	 * EDIT New API available, see Playground.SetWantedSpectatorCameraType(...)
	 */
     // Playground.SetWantedSpectatorCameraType(CPlaygroundClient::ESpectatorCameraType::Follow); // SetWantedSpectatorCameraType changes "desired" camera, which means player can still change it with inputs (instead of "forced")
void SetSpectatorTarget(bool visible) {
    if (!visible) {
        PlaygroundScript.UIManager.UIAll.SpectatorForceCameraType = 2;
        // Spectator_SetForcedTarget_Clear();
        // nope:
        // PlaygroundScript.UIManager.UIAll.Spectator_SetForcedTarget_AllPlayers();
        // sort of decent if you can find the finish checkpoint:
        // PlaygroundScript.UIManager.UIAll.Spectator_SetForcedTarget_Landmark(CurrentPlayground.Arena.MapLandmarks[0]);
    } else {
        // Spectator_SetForcedTarget_Clear();
        Spectator_SetForcedTarget_Ghost(forcedTarget);
        PlaygroundScript.UIManager.UIAll.SpectatorForceCameraType = 1;
        // PlaygroundScript.UIManager.UIAll.SpectatorForceCameraType = 15; // disable?
        // Network.PlaygroundClientScriptAPI.SetWantedSpectatorCameraType(CGamePlaygroundClientScriptAPI::ESpectatorCameraType::Follow);
    }
}

void ResetSpectatorRace() {
    // When following records, reset ghosts when scrubber reaches end (unless paused/scrubbing)
    if (IsFollowingRecord() && !scrubber.Paused && !slider.IsMouseDown && PlaygroundScript !is null) {
        scrubber.lastStart = PlaygroundScript.Now;
        Ghosts_SetStartTime(scrubber.lastStart);
    }
}

/*
 detection:

 Keep track of latest start time passed to Ghosts_SetStartTime, since this is
 called before Spectator_SetForcedTarget_Ghost to set proper initial start time

 decide that spectating a records ghost when:
 Spectator_SetForcedTarget_Ghost() called if SpectatorForceCameraType == 1

 things to mess with when spectating:
 - Block Ghosts_SetStartTime() and try to manage that here
 - Block fade out animations since they won't happen at good times after trying
   to manage Ghosts_SetStartTime, looks a little worse but fine-ish
 - Script can change SpectatorForceCameraType at a bad time

 decide that spectating is done when:
 UISequence_Current is no longer EndRound (2023/2/3 or UIInteraction for View Records widget)

 Has side effect of not restarting playback when clicking a different name in
 the records widget, maybe preferable since you can just scrub to start if
 important. Similarly the poor end detection can be alleviated by scrubbing, not
 great but ehhhhh

*/

// Intercept:

// void Ghosts_SetStartTime(int StartTime)
bool Intercept_Ghosts_SetStartTime = true;
void Ghosts_SetStartTime(int startTime) {
    Intercept_Ghosts_SetStartTime = false;
    if (PlaygroundScript !is null) {
        PlaygroundScript.Ghosts_SetStartTime(startTime);
    }
    Intercept_Ghosts_SetStartTime = true;
}

bool int_CSmArenaRulesMode_Ghosts_SetStartTime(CMwStack&in stack) {
    if (!Setting_Ghost_RecordsSpectate) { return true; }

    if (Intercept_Ghosts_SetStartTime) {
        uint startTime = stack.CurrentInt(0);
        lastStartTime = startTime;

        log("Ghosts_SetStartTime(" + startTime + ")" + " recordsFollow=" + recordsFollow);

        // ignore calls whhen trying to control it
        if (recordsFollow) {
            return false;
        }

        return true;
    }

    return true;
}

// void FlushAndAddFromXml(CGameManialinkControl@ Control, string XmlTarget, uint Duration, CGameManialinkAnimManager::EAnimManagerEasing EasingFunc)
bool int_CGameManialinkAnimManager_FlushAndAddFromXml(CMwStack&in stack) {
    if (!Setting_Ghost_RecordsSpectate || !recordsFollow) {
        return true;
    }

    auto Control = cast<CGameManialinkControl@>(stack.CurrentNod(3));

    // see Fade_Client.Script.txt
    if (Control !is null && Control.ControlId == "quad-fade") {
            string XmlTarget = stack.CurrentString(2);
            // uint Duration = stack.CurrentUint(1);
            // auto Easing = CGameManialinkAnimManager::EAnimManagerEasing(stack.CurrentEnum(0));

            bool isFadeIn = XmlTarget == '<a opacity="0" />';
            // log("Fade (fadein? " + isFadeIn + "): AnimMgr.Add(" + Control.ControlId + ", " + XmlTarget + ", " + Duration + ", " + tostring(Easing));
            return isFadeIn; // always allow fade-ins in case state is messed up
    }

    return true;
}

bool Intercept_Spectator_SetForcedTarget_Ghost = true;
void Spectator_SetForcedTarget_Ghost(MwId&in id) {
    Intercept_Spectator_SetForcedTarget_Ghost = false;
    if (PlaygroundScript !is null && PlaygroundScript.UIManager !is null) {
        PlaygroundScript.UIManager.UIAll.Spectator_SetForcedTarget_Ghost(id);
    }
    Intercept_Spectator_SetForcedTarget_Ghost = true;
}

// void Spectator_SetForcedTarget_Ghost(MwId GhostInstanceId)
bool int_CGamePlaygroundUIConfig_Spectator_SetForcedTarget_Ghost(CMwStack&in stack) {
    if (Intercept_Spectator_SetForcedTarget_Ghost) {
        forcedTarget = stack.CurrentId(0);

        // want to keep track of the forced target regardless
        if (!Setting_Ghost_RecordsSpectate) { return true; }

        log('Spectator_SetForcedTarget_Ghost(' + forcedTarget.GetName() + ')');

        if (PlaygroundScript !is null
            && PlaygroundScript.UIManager !is null
            // iffy:
            && PlaygroundScript.UIManager.UIAll.SpectatorForceCameraType == 1) {
            StartRecordsSpectate();
        }
    }

    return true;
}

void Spectator_SetForcedTarget_Clear() {
    if (PlaygroundScript !is null
        && PlaygroundScript.UIManager !is null) {
        PlaygroundScript.UIManager.UIAll.Spectator_SetForcedTarget_Clear();
    }
}

// void Spectator_SetForcedTarget_Clear()
// bool int_CGamePlaygroundUIConfig_Spectator_SetForcedTarget_Clear(CMwStack&in stack) {
//     if (!Setting_Ghost_RecordsSpectate) { return true; }
//     forcedTarget = MwId(uint(-1));
//     log("SetForcedTarget_Clear()");
//     return true;
// }

// Main/Cleanup:

void Cleanup() {
    ClearTemporaryGhosts();
    Ghosts_SetStartTime(-1);
}

void OnDestroyed() {
    Cleanup();
}

void OnDisabled() {
    Cleanup();
}

void Main() {
    @scrubber = GhostScrubber();
    @slider = IGSlider();

    @scrubber.OnReset = ClearTemporaryGhosts;

    slider.ValueFormat = "";
    slider.ButtonText = SCRUBBER_PLAY_TEXT;

    @slider.OnChange = OnSliderChange;

    Dev::InterceptProc("CGameManialinkAnimManager", "FlushAndAddFromXml", int_CGameManialinkAnimManager_FlushAndAddFromXml);
    Dev::InterceptProc("CSmArenaRulesMode", "Ghosts_SetStartTime", int_CSmArenaRulesMode_Ghosts_SetStartTime);
    Dev::InterceptProc("CGamePlaygroundUIConfig", "Spectator_SetForcedTarget_Ghost", int_CGamePlaygroundUIConfig_Spectator_SetForcedTarget_Ghost);
    // Dev::InterceptProc("CGamePlaygroundUIConfig", "Spectator_SetForcedTarget_Clear", int_CGamePlaygroundUIConfig_Spectator_SetForcedTarget_Clear);
}
