
funcdef void RaceReset();

class GhostScrubber {
    int RaceTime {
        get {
            return Math::Clamp(relativeRaceTime, 0, RaceLength);
        }
    }

    // FIXME: bad
    int RaceLength {
        get {
            return Math::Max(autoRaceLength, manualRaceLength);
        }
        set {
            manualRaceLength = value;
        }
    }

    private int manualRaceLength = 0;
    private int autoRaceLength = 0;

    int lastStart = 0;
    private int lastRaceTime = 0;
    private int relativeRaceTime = 0;
    int PauseRaceTime = 0;
    private uint64 unpauseTime = 0;
    private bool dirty = false;

    void Reset() {
        ResetTime();
        lastStart = lastRaceTime = relativeRaceTime = 0;
        if (Paused) { Paused = false; }
    }

    RaceReset@ OnReset = null;

    void ResetTime() {
        // however if reset in end of round replay, the replay will stop, so
        // don't do that
        if (dirty
            && GameTerminal !is null
            && GameTerminal.UISequence_Current == CGamePlaygroundUIConfig::EUISequence::Playing) {
            log('reset ghost start');
            if (PlaygroundScript !is null) {
                Ghosts_SetStartTime(-1);
                auto player = ControlledPlayer;
                if (player is null) { return; }
                lastStart = PlaygroundScript.Now - cast<CSmScriptPlayer@>(player.ScriptAPI).CurrentRaceTime;
            }
            if (OnReset !is null) {
                OnReset();
            }
            dirty = false;
        }
    }

    bool _paused;
    bool Paused {
        get { return _paused; }
        set {
            _paused = value;
            PauseRaceTime = value ? relativeRaceTime : 0;
            unpauseTime = 0;
        }
    }

    void PauseFor(int time) {
        if (time < 0) { return; }
        Paused = true;
        unpauseTime = Time::Now + time;
    }

    int GetRaceLength() {
        int raceLength = 0;
        if (PlaygroundScript !is null && PlaygroundScript.DataFileMgr !is null) {
            auto DataFileMgr = PlaygroundScript.DataFileMgr;
            for (uint i = 0; i < DataFileMgr.Ghosts.Length; i++) {
                auto ghost = DataFileMgr.Ghosts[i];
                raceLength = Math::Max(ghost.Result.Time, raceLength);
            }
        }
        if (Network !is null && Network.ClientManiaAppPlayground !is null && Network.ClientManiaAppPlayground.DataFileMgr !is null) {
            auto DataFileMgr = Network.ClientManiaAppPlayground.DataFileMgr;
            for (uint i = 0; i < DataFileMgr.Ghosts.Length; i++) {
                auto ghost = DataFileMgr.Ghosts[i];
                raceLength = Math::Max(ghost.Result.Time, raceLength);
            }
        }
        return raceLength;
    }

    void UpdateRaceTime() {
        relativeRaceTime = PlaygroundScript.Now - lastStart;
    }

    void SetRaceTime(int newRaceTime) {
        if (newRaceTime >= RaceLength) {
            return;
        }
        if (PlaygroundScript is null) { return; }
        int newStartTime = PlaygroundScript.Now - newRaceTime;
        Ghosts_SetStartTime(newStartTime);
        dirty = true;
        lastStart = newStartTime;
        UpdateRaceTime();
        if (Paused) {
            PauseRaceTime = RaceTime;
        }
    }

    int lastUpdate = 0;

    void Update(float dt) {
        if (PlaygroundScript is null) { return; }

        // Update time
        autoRaceLength = GetRaceLength();

        // quick state detection

        auto player = ControlledPlayer;
        if (player !is null) {
            auto raceTime = cast<CSmScriptPlayer@>(player.ScriptAPI).CurrentRaceTime;
            if (raceTime < 0 && lastRaceTime > raceTime) {
                // reset ghost start time in countdown
                Reset();
                // set start time when race starts
                lastStart = PlaygroundScript.Now - raceTime;
            }
            lastRaceTime = raceTime;
        }

        UpdateRaceTime();

        if (Paused) {
            SetRaceTime(PauseRaceTime);
            if (unpauseTime != 0 && unpauseTime < Time::Now) {
                Paused = false;
            }
        }
    }
}
